"""
    ZDataCombiner
    =============

Allow multiple datasources to be combined (In SQL Terms 'joined') to form
a single datasource.
"""
# $Id: ZDataCombiner.py,v 1.3 2002/01/13 21:08:58 adrian Exp $
#
# $Log: ZDataCombiner.py,v $
# Revision 1.3  2002/01/13 21:08:58  adrian
# Completed relational logic in ZDataCombiner.
# Fixed up compatibility bugs in combiner and report.
# Started online help pages.
#
# Revision 1.2  2001/10/21 15:03:36  adrian
# no message
#
#
__version__ = '$Revision: 1.3 $'[11:-2]

import Globals
import urllib
import OFS.SimpleItem
import OFS.PropertyManager
import OFS.ObjectManager
import Persistence
import AccessControl.Role
import Acquisition
import DataSourceReader
import ZVisualQuery
import DateTime
import string
from DocumentTemplate import DT_HTML
import sys
import Shared.DC.ZRDB.Results

manage_addZDataCombinerForm = Globals.HTMLFile('www/addZDataCombiner', globals())

def manage_addZDataCombiner(self, id, title='', submit='', REQUEST=None):
    """ """
    obj = ZDataCombiner(id, title=title)
    self._setObject(obj.getId(), obj)
    if REQUEST is not None:
        try: u = self.DestinationURL()
        except: u = REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u, urllib.quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_workspace')
    return ''

class ZDataCombiner(
        OFS.PropertyManager.PropertyManager,
        OFS.SimpleItem.SimpleItem,
        Acquisition.Implicit,
        Persistence.Persistent,
        AccessControl.Role.RoleManager,
#        RelLogic.RelLogic,
    ):
    """ """
    meta_type = 'Z Data Combiner'
    id = 'ZDataCombiner'
    index_html = None

    # Masquerade as function:
    class func_code: pass
    func_code=func_code()
    func_code.co_varnames='self','REQUEST','RESPONSE'
    func_code.co_argcount=3

    manage_combinerOptions = Globals.HTMLFile('www/combinerOptions', globals())
    manage_Joins = Globals.HTMLFile('www/queryJoins', globals())
    manage_OutputColumns = Globals.HTMLFile('www/queryOutputColumns', globals())
    manage_options = (
        (
            {'label': 'Combiner Options', 'action': 'manage_combinerOptions',
             'help': ('ZDataQueryKit', 'CombinerOptions.stx'), },
        ) +
        OFS.PropertyManager.PropertyManager.manage_options +
        (
            {'label': 'Joins', 'action': 'manage_Joins',
             'help': ('ZDataQueryKit', 'QueryJoins.stx'), },
            {'label': 'Output Columns', 'action': 'manage_OutputColumns',
             'help': ('ZDataQueryKit', 'QueryColumns.stx'), },
            {'label': 'Preview', 'action': 'manage_preview',
             'help': ('ZDataQueryKit', 'CombinerView.stx'), },
        ) +
        AccessControl.Role.RoleManager.manage_options +
        OFS.SimpleItem.SimpleItem.manage_options
    )
    _properties	= (
        {'id': 'title', 'type': 'string', 'mode': 'w'},
    )
    relationActions = (
        '=',
        '!=',
        '>',
        '<',
        '>=',
        '<=',
    )
    datasources = {}
    joins = []
    columnCache = []
    columns = {}
    class_name_ = ""
    class_file_ = ""
    _v_brain = None

    def __init__(self, id, title=''):
        """ """
        self.id = id
        self.title = title

    def manage_preview(self, REQUEST=None, RESPONSE=None):
        """Returns an HTML representation of the OSS's data"""
        dtml = "<dtml-var manage_page_header><dtml-var manage_tabs>"
        if not self.datasources:
            dtml = dtml + '<p class="form-help">Before you can preview this data combiner, you must select some datasources to combine</p>'
        else:
            dtml = dtml + "<br/><table border=1><tr><th>"
            colNames = self.names()
            dtml = dtml + string.join(colNames, "</th>\n<th>") + "</th></tr>\n"
            rows = self._rows()
            for row in rows:
                dtml = dtml + "<tr>"
                for col in colNames:
                    dtml = dtml + "<td>%s</td>" % (row[col], )
                dtml = dtml + "</tr>\n"
            dtml = dtml + "</table><br>\n"
            dtml = dtml + "%d records selected" % (len(rows), )
        dtml = dtml + "<dtml-var manage_page_footer>"
        obj = DT_HTML.HTML(dtml, self.REQUEST,
            'XMLODBCClient - manage_preview - %s' % self.getId())
        return obj(self.aq_inner, REQUEST)

    def manage_addDataSource(self, datasource, REQUEST=None):
        """ Add a data source to the mix """
        self._addDataSource(datasource)
        if REQUEST:
            message = "Datasource added."
            return self.manage_combinerOptions(self, REQUEST, manage_tabs_message=message)

    def _addDataSource(self, dsID):
        """ Add a new datasource """
        newDS = None
        if hasattr(self, dsID) or hasattr(dsID, 'getId'):
            try:
                newDS = DataSourceReader.DataSourceReader(acquirer=self, datasource=dsID)
                dsID = newDS.DSR_data_source_id
            except:
                newDS = None
                dsID = None
        if dsID and newDS:
            ds = self.datasources
            ds[dsID] = newDS
            self.datasources = ds
            clist = self.columnCache
            for column in self.datasources[dsID]._getColumnList():
                clist.append((dsID, column))
            self.columnCache = clist

    def manage_removeDataSources(self, REQUEST):
        """ Remove a data source from the mix """
        list = self.datasources.keys()
        for item in list:
            if REQUEST.has_key('ds_%s' % item):
                self._removeDataSource(item)
        if REQUEST:
            message = "Datasource removed."
            return self.manage_combinerOptions(self, REQUEST, manage_tabs_message=message)

    def _removeDataSource(self, dsID):
        """ Remove a single datasource """
        chash = self.columns
        for column in self.columns.keys():
            if self._getTableName(column) == dsID:
                del chash[column]
        self.columns = chash
        clist = []
        for column in self.columnCache:
            if not column[0] == dsID:
                clist.append(column)
        self.columnCache = clist
        joins = []
        for item in self.joins:
            if dsID not in (self._getTableName(item['leftColumn']), self._getTableName(item['rightColumn'])):
                joins.append(item)
        self.joins = joins
        ds = self.datasources
        del ds[dsID]
        self.datasources = ds

    def manage_recacheDataSources(self, REQUEST):
        """ Reload the columnCache, and clear the output columns """
        self.columns = {}
        list = self.datasources.keys()
        for item in list:
            self._removeDataSource(item)
            self._addDataSource(item)
        if REQUEST:
            message = "Column caches updated."
            return self.manage_combinerOptions(self, REQUEST, manage_tabs_message=message)

    def manage_changeOutputColumns(self, REQUEST):
        """ """
        chash = {}
        for column in self.columnCache:
            sqlName = "%s.%s" % (column[0], column[1])
            cbFormName = "c_%s_%s" % (column[0], column[1])
            txtFormName = "a_%s_%s" % (column[0], column[1])
            if REQUEST.get(cbFormName, ''):
                chash[sqlName] = REQUEST.get(txtFormName)
        self.columns = chash
        if REQUEST:
            message = "Output columns updated."
            return self.manage_OutputColumns(self, REQUEST, manage_tabs_message=message)

    def manage_addJoin(self, leftColumn, relation, rightColumn, REQUEST=None):
        """ Add a new table join """
        joins = self.joins
        joins.append({'leftColumn': str(leftColumn), 'relation': str(relation), 'rightColumn': str(rightColumn)})
        self.joins = joins
        if REQUEST:
            message = "Join created."
            return self.manage_Joins(self, REQUEST, manage_tabs_message=message)

    def manage_delJoin(self, REQUEST=None):
        """ Delete one or more table joins """
        joins = self.joins
        for item in range(len(self.joins)-1, -1, -1):
            if REQUEST.get('dl_%d' % item):
                del joins[item]
        self.joins = joins
        if REQUEST:
            message = "Join(s) deleted."
            return self.manage_Joins(self, REQUEST, manage_tabs_message=message)

    def tables(self):
        """ A hack to allow the OutputColumns from VQD to work """
        return self.datasources.keys()

    def _getTableName(self, columnName):
        """ Split the table name from a FQ column name """
        if string.find(columnName, '.')==-1:
            raise "Not Fully Qualified Column Name"
        return columnName[:string.find(columnName, '.')]

    def _getColumnName(self, columnName):
        """ Split the column name from a FQ column name """
        if string.find(columnName, '.')==-1:
            raise "Not Fully Qualified Column Name"
        return columnName[string.find(columnName, '.')+1:]

    def names(self):
        """ Get the names of the (output) columns in the mix """
        list = []
        for item in self.columns.keys():
            list.append(self.columns[item])
        return tuple(list)

    def _data_dictionary(self):
        """ Get the data dictionary for the mix """
        dd = []
        for item in self.columns.keys():
            dsID = self._getTableName(item)
            column = self._getColumnName(item)
            DDE = self.datasources[dsID]._getDataDictionary()[column]
            dd.append({'name': '%s' % self.columns[item], 'type': '%s' % DDE['type']})
        return tuple(dd)

    def _leftOnly(self):
        """ List the tables that appear only on the
            left hand side of a join, or appear in
            no joins at all """
        tabs = {}
        for item in self.tables:
            tabs[item] = item
        for item in self.joins:
            rtab = item['rightColumn']
            rtab = self._getTableName(rtab)
            if tabs.has_key(rtab):
                del tabs[rtab]
        if not tabs:
            raise "Product Error", "At least one table must not appear on the right"
        return tabs.keys()

    def _combine_rows(self, rowL, prefixL, rowR, prefixR):
        """ Combine the fields from rowL with
            those from rowR. """
        out = {}
        for fieldL in rowL.keys():
            if prefixL:
                out["%s.%s" % (prefixL, fieldL)] = rowL[fieldL]
            else:
                out[fieldL] = rowL[fieldL]
        for fieldR in rowR.keys():
            if prefixR:
                out["%s.%s" % (prefixR, fieldR)] = rowR[fieldR]
            else:
                out[fieldR] = rowR[fieldR]
        return out

    def _product(self, dsL, prefixL, dsR, prefixR, joins=[]):
        """ Generate the product of dsL with dsR
            adding to the output only those rows that
            match the joins. DSL is either the table
            mentioned in joins[n]['leftColumn'] or
            an existing product the contains it. """
        out = []
        for rowL in dsL:
            for rowR in dsR:
                row = self._combine_rows(rowL, prefixL, rowR, prefixR)
# Test join constraints
                include = 1
                for join in joins:
                    leftValue = row[join['leftColumn']]
                    rightValue = row[join['rightColumn']]
                    rel = join['relation']
                    if rel == '=':
                        if not leftValue == rightValue: include = 0
                    elif rel == '!=':
                        if not leftValue != rightValue: include = 0
                    elif rel == '>':
                        if not leftValue > rightValue: include = 0
                    elif rel == '<':
                        if not leftValue < rightValue: include = 0
                    elif rel == '>=':
                        if not leftValue >= rightValue: include = 0
                    elif rel == '<=':
                        if not leftValue <= rightValue: include = 0
                if include:
                    out.append(row)
        return out

    def _match_joins(self, dsList1, dsList2):
        """ find all the joins referencing something
            from list1 and something from list2 """
        out = []
        for join in self.joins:
            dsL = self._getTableName(join['leftColumn'])
            dsR = self._getTableName(join['rightColumn'])
            if (dsL in dsList1 and dsR in dsList2) or (dsL in dsList2 and dsR in dsList1):
                out.append(join)
        return out

    def _rows(self):
        """ Return the data rows """
        rows = []
        if not self.datasources:
            return tuple(rows)
# compute row product (Scarey - I hope the user defined some joins!)
        dsList = self.datasources.keys()
        dsDoneList = [dsList[0]]
        for ds in dsList[1:]:
            myjoins = self._match_joins(dsDoneList, [ds])
            if len(dsDoneList)==1:
                rows = self._product(self.datasources[dsList[0]]._getData(self), dsList[0], self.datasources[ds]._getData(self), ds, myjoins)
            else:
                rows = self._product(rows, None, self.datasources[ds]._getData(self), ds, myjoins)
            dsDoneList.append(ds)
# Now remove unwanted columns, and rename wanted ones
        out = []
        outColumns = self.columns.keys()
        for row in rows:
            newRow = {}
            for field in row.keys():
                if field in outColumns:
                    newRow[self.columns[field]] = row[field]
            out.append(newRow)
        return tuple(out)

    def _results(self, rows):
        """ """
        if hasattr(self, '_v_brain'):
            brain = self._v_brain
        else:
            brain = self._v_brain = getBrain(self.class_file_, self.class_name_)
        return Shared.DC.ZRDB.Results.Results((self._data_dictionary(), rows), brains=brain, parent=None)

    def __call__(self, *args, **kargs):
        """ """
        return self._results(self._rows())
