"""
    ZReportTool
    ===========

This product is intended to take a list of dictionaries as its input,
and output an nicely formatted HTML table.

Done:
* Select input source
* Select column names
* Order column names
* Select "group by" columns
* Select "sum on" columns
* Select CSS Classes for rows
* Add calculated columns
* Produce output table
* Add __call__ method so that ZReportTools can be dropped into other objects/pages
"""
# $Id: ZReportTool.py,v 1.5 2002/01/13 21:08:58 adrian Exp $
#
# $Log: ZReportTool.py,v $
# Revision 1.5  2002/01/13 21:08:58  adrian
# Completed relational logic in ZDataCombiner.
# Fixed up compatibility bugs in combiner and report.
# Started online help pages.
#
# Revision 1.4  2002/01/13 09:36:32  adrian
# Fixed bug in ZReportTool that prevented object creation
#
# Revision 1.3  2001/10/29 09:29:56  adrian
# Cleaned up a number of "Ambiguous name..." warnings
#
# Revision 1.2  2001/10/21 15:03:36  adrian
# no message
#
#
__version__ = '$Revision: 1.5 $'[11:-2]

# Todo
#   Change calculated columns not to use eval()

import urllib
import Globals
import OFS.SimpleItem
import OFS.PropertyManager
import AccessControl.Role
import App.Management
import Acquisition
import Persistence

import DataSourceReader

manage_addZReportToolForm = Globals.HTMLFile('www/addZReportTool', globals())

def manage_addZReportTool(self, id, title='', report_class='report', heading_class='heading', data_class='data', grand_total_class='grand_total', data_source='', submit='', REQUEST=None):
    """ """
    if data_source:
        data = getattr(self, data_source)
    else:
        data = ''
    obj = ZReportTool(str(id),
        title=str(title),
        report_class=str(report_class),
        heading_class=str(heading_class),
        data_class=str(data_class),
        grand_total_class=str(grand_total_class),
        data_source=data,
    )
    self._setObject(obj.getId(), obj)
    if REQUEST is not None:
        try: u = self.DestinationURL()
        except: u = REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u, urllib.quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_workspace')
    return ''

class ZReportTool(
        OFS.PropertyManager.PropertyManager,
        OFS.SimpleItem.SimpleItem,
        Acquisition.Implicit,
        Persistence.Persistent,
        AccessControl.Role.RoleManager,
        DataSourceReader.DataSourceReader,
        ):
    """ Produce an HTML formatted report from a datasource """

    meta_type = 'Z Report Tool'
    id = 'ZReportTool'
    manage_main = Globals.HTMLFile('www/reportOptions', globals())
    manage_options  = (
        (
            {'label': 'Column Options', 'action': 'manage_main',
             'help': ('ZDataQueryKit', 'ReportOptions.stx')},
        ) +
        OFS.PropertyManager.PropertyManager.manage_options +
        (
            {'label': 'View', 'action': 'index_html'},
        ) +
        AccessControl.Role.RoleManager.manage_options +
        OFS.SimpleItem.SimpleItem.manage_options
    )
    _properties	=   (
        {'id': 'title', 'type': 'string', 'mode': 'w'},
        {'id': 'report_class', 'type': 'string', 'mode': 'w'},
        {'id': 'heading_class', 'type': 'string', 'mode': 'w'},
        {'id': 'data_class', 'type': 'string', 'mode': 'w'},
        {'id': 'grand_total_class', 'type': 'string', 'mode': 'w'},
    )
    
    def __init__(self, id, title='', report_class='report', heading_class='heading', data_class='data', grand_total_class='grand_total', data_source=''):
        """ """
        self.id = id
        self.title = title
        self.report_class = report_class
        self.heading_class = heading_class
        self.data_class = data_class
        self.grand_total_class = grand_total_class
        self.manage_changeDataSource(data_source)

    def manage_changeDataSource(self, data_source, REQUEST=None):
        """ """
# Contains a list of column names in display order
        self.columnOrder = ()
# Contains either {'name': ..., 'type': ...} for database columns
# or {'name': ..., 'type': 'calc', 'formular': ... } for calculated columns
        self.columnData = {}
# Contains the printing titles of each column
        self.columnTitle = {}
# Contains the alignment of each column
        self.columnAlign = {}
# Contains the CSS class of the heading and subtotal/footer rows grouped on this column
        self.columnClass = {}
# 0 = Hide, 1 = Normal, 2 = Normal w/ Sub-Totals, 3 = Group, 4 = Group w/ Summary only
        self.flagGroup = {}
# A tuple containing the grouping columns in order
        self.groupColumns = ()
# A tuple containing the totals columns in order
        self.totalColumns = ()
# A tuple containing the calculated columns in order
        self.calcColumns = ()
# The number of visible columns
        self.displayCount = 0
        try:
            self._setDataSource(datasource=data_source)
            self.columnOrder = self._getColumnList()
# Ensure that this is a deep copy
            dd = self._getDataDictionary()
            for columnName in self.columnOrder:
                self.columnData[columnName] = {'name': '%s' % dd[columnName]['name'], 'type': '%s' % dd[columnName]['type']}
                self.flagGroup[columnName] = '1'
                self.columnTitle[columnName] = columnName
                self.columnAlign[columnName] = 'LEFT'
                if self.columnData[columnName]['type'] == 'i' or self.columnData[columnName]['type'] == 'l' or self.columnData[columnName]['type'] == 'n':
                    self.columnAlign[columnName] = 'RIGHT'
                self.columnClass[columnName] = columnName
            return self.manage_saveDataOptions(REQUEST=REQUEST, submit='Move First')
        except:
            raise
            if REQUEST:
                message="Data Definitions cleared."
                return self.manage_main(self, REQUEST, manage_tabs_message=message)

    def manage_saveDataOptions(self, REQUEST=None, manage_saveDataOptions=None):
        """ """
        submit = manage_saveDataOptions
        cols1 = []
        cols2 = []
        for columnName in self.columnOrder:
            if submit == 'Move First':
                if REQUEST and REQUEST.get('f_%s' % columnName, ''):
                    cols1.append(columnName)
                else:
                    cols2.append(columnName)
            elif submit == 'Delete':
                if not REQUEST.get('f_%s' % columnName, ''):
                    cols1.append(columnName)
                    del self.columnTitle[columnName]
                    del self.columnName[columnName]
                    del self.columnClass[columnName]
                    del self.columnData[columnName]
                    del self.flagGroup[columnName]
            elif submit == 'Save Changes':
                self.columnTitle[columnName] = REQUEST.get('l_%s' % columnName, columnName)
                self.columnAlign[columnName] = REQUEST.get('a_%s' % columnName, 'LEFT')
                self.columnClass[columnName] = REQUEST.get('c_%s' % columnName, columnName)
                if self.columnData[columnName]['type'] == 'calc':
                    self.columnData[columnName]['formular'] = REQUEST.get('cf_%s' % columnName, "'N/A'")
                else:
                    self.flagGroup[columnName] = REQUEST.get('g_%s' % columnName, '1')
            else:
                raise "Parameter error", "Don't know how to %s on this form" % submit
        if submit in ['Move First', 'Delete']:
            self.columnOrder = tuple(cols1 + cols2)
        self.groupColumns = []
        self.totalColumns = []
        self.calcColumns = []
        self.displayCount = 0
        for columnName in self.columnOrder:
            if self.flagGroup[columnName] in ['3', '4', '5']:
                self.groupColumns.append(columnName)
            if self.flagGroup[columnName] in ['2']:
                self.totalColumns.append(columnName)
            if self.columnData[columnName]['type'] == 'calc':
                self.calcColumns.append(columnName)
            if not self.flagGroup[columnName] in ['0', '3']:
                self.displayCount = self.displayCount + 1
        self.groupColumns = tuple(self.groupColumns)
        self.totalColumns = tuple(self.totalColumns)
        self.calcColumns = tuple(self.calcColumns)

# Return success to the browser
        if REQUEST:
            message="Data Definitions Updated."
            return self.manage_main(self, REQUEST, manage_tabs_message=message)

    def manage_addColumn(self, columnName='', columnTitle='', columnFormular='', REQUEST=None):
        """ Add a new calculated column to every record """
        if columnName in self.columnOrder:
            raise "Column already exists", "The requested column name already exists in this dataset"
        cols = map(None, self.columnOrder)
        cols.append(columnName)
        self.columnOrder = tuple(cols)
        self.flagGroup[columnName] = '1'
        self.columnData[columnName] = {'name': columnName, 'type': 'calc', 'formular': columnFormular}
        self.columnTitle[columnName] = columnTitle
        self.columnAlign[columnName] = 'LEFT'
        self.columnClass[columnName] = columnName
        if REQUEST:
            message="Column Added."
            return self.manage_main(self,REQUEST, manage_tabs_message=message)

    def _record(self, record, CSS):
        """ Take a database record and output a record
            containing only the display columns """
# At the moment I use eval() to calculate the formular of a calculated column
# This is very-unsafe, so I will need to dig out the one dtml-var (etc) uses
        rec = {'class': CSS}
        for columnName in self.columnOrder:
            try:
                if self.columnData[columnName]['type'] == 'calc':
                    rec[columnName] = eval(self.columnData[columnName]['formular'], record)
                else:
                    rec[columnName] = record[columnName]
            except:
                rec[columnName] = ""
        return rec

    def _chewGroup(self, branch, groups):
        """ Recurse down the dataTree, moving data rows to the
            detail branches and creating totals rows """
        if not groups:
            return
        this_group = groups[0]
        remaining_groups = groups[1:]
        rows = map(None, branch['~ rows'])
        branch['~ rows'] = []
        for record in rows:
            this_value = record[this_group]
            if not branch.has_key(this_value):
                branch[this_value] = {}
            if not branch[this_value].has_key('~ rows'):
                branch[this_value]['~ rows'] = []
            branch[this_value]['~ rows'].append(record)
        for this_value in branch.keys():
            if this_value != '~ rows':
                self._chewGroup(branch[this_value], remaining_groups)

# Node total for leaf node == total of rows
# Node total for non-leaf node == total of child-totals
    def _makeTotals(self, group, branch, groups):
        """ Add up the totals in this branch """
        rList = []
        tRec = {}
        kList = branch.keys()
        if branch.has_key('~ rows'):
            kList.remove('~ rows')
            rList = branch['~ rows']
        if branch.has_key('~ total'):
            kList.remove('~ total')
            tRec = branch['~ total']
        kList.sort()
        for subBranch in kList:
            self._makeTotals(groups[0], branch[subBranch], groups[1:])

        newRecord = {}
        if group:
            css = self.columnClass[group]
        else:
            css = self.grand_total_class
        for columnName in self.totalColumns:
            newRecord[columnName] = 0
        if kList:
# Total child-totals
            for subBranch in kList:
                if branch[subBranch].has_key('~ total'):
                    record = branch[subBranch]['~ total']
                    for columnName in self.totalColumns:
                        newRecord[columnName] = newRecord[columnName] + record[columnName]
        else:
# Total of rows
            for record in branch['~ rows']:
                for columnName in self.totalColumns:
                    newRecord[columnName] = newRecord[columnName] + record[columnName]
        branch['~ total'] = self._record(newRecord, css)

    def _chewData(self):
        """ Return the dataTree """
        dataTree = {}
        dataTree['~ rows'] = []
        data = self._getData()
        for record in data:
            dataTree['~ rows'].append(self._record(record, self.data_class))
        self._chewGroup(dataTree, self.groupColumns)
        self._makeTotals('', dataTree, self.groupColumns)
        return dataTree

    def _Header(self):
        """ Print the table header row """
        ret = "<tr>\n"
        for columnName in self.columnOrder:
            if not self.flagGroup[columnName] in ['0', '3']:
                ret = ret + '<th align="' + self.columnAlign[columnName] + '" class="' + self.heading_class + '">' + self.columnTitle[columnName] + '</th>\n'
        ret = ret + '</tr>\n'
        return ret

    def _Field(self, Record, columnName):
        """ Format field according to datatype and content """
        str = '<td align="' + self.columnAlign[columnName] + '" class="' + Record['class'] + '">'
        if Record.has_key(columnName):
            value = Record[columnName]
            dataType = self.columnData[columnName]['type']
            format = '%s'
            if dataType == 'i':
                format = '%d'
            elif dataType == 'l' or dataType == 'n':
                format = '%.2f'
            if value:
                str = str + format % value
            else:
                str = str + '&nbsp;'
        else:
            str = str + '&nbsp;'
        str = str + '</td>\n'
        return str

    def _Row(self, Record):
        """ Display a formatted row of data """
        str = '<tr>\n'
        for columnName in self.columnOrder:
            if not self.flagGroup[columnName] in ['0', '3']:
                str = str + self._Field(Record, columnName)
        str = str + '</tr>\n'
        return str

    def _GroupHeader(self, prefix, thisGroup, thisValue):
        """ Display a header for this group. """
        if thisGroup and self.flagGroup[thisGroup] != '5':
            css = self.columnClass[thisGroup]
        else:
            return ''
        if not thisValue:
            thisValue = '[None]'
        str = '<tr>\n'
        str = str + '<td align="LEFT" colspan="%d" class="%s">%s %s</td>\n' % (self.displayCount, css, prefix, thisValue)
        str = str + '</tr>\n'
        return str

    def _GroupFooter(self, prefix, thisGroup, thisValue):
        """ Display a footer for this group. """
        str = '<tr>\n'
        if not thisValue:
            thisValue = '[None]'
        if thisGroup:
            if self.totalColumns:
                str = str + '<td align="LEFT" colspan="%d" class="%s">Totals for %s %s</td>\n' % (self.displayCount, self.columnClass[thisGroup], prefix, thisValue)
            else:
                str = str + '<td align="LEFT" colspan="%d" class="%s">End of %s %s</td>\n' % (self.displayCount, self.columnClass[thisGroup], prefix, thisValue)
        else:
            if self.totalColumns:
                str = str + '<td align="LEFT" colspan="%d" class="%s">Grand Totals</td>\n' % (self.displayCount, self.grand_total_class)
            else:
                return ''
        str = str + '</tr>\n'
        return str

    def _DescendRows(self, branch, prefix, thisGroup, thisValue, groups):
        """ """
        retStr = self._GroupHeader(prefix, thisGroup, thisValue)
        rList = []
        tRec = {}
        kList = branch.keys()
        if branch.has_key('~ rows'):
            kList.remove('~ rows')
            rList = branch['~ rows']
        if branch.has_key('~ total'):
            kList.remove('~ total')
            tRec = branch['~ total']
        kList.sort()
        if thisGroup == '' or self.flagGroup[thisGroup] != '5':
            for k in kList:
                retStr = retStr + self._DescendRows(branch=branch[k], prefix="%s %s" % (prefix, thisValue), thisGroup=groups[0], thisValue=k, groups=groups[1:])
            for Record in rList:
                retStr = retStr + self._Row(Record)
        retStr = retStr + self._GroupFooter(prefix, thisGroup, thisValue)
        if self.totalColumns:
            retStr = retStr + self._Row(tRec)
        return retStr

    def __call__(self, REQUEST=None):
        """ Call the report from a webpage (Or anywhere). """
# Setup a temporary data-tree
        dataTree = self._chewData()

# Print the data from the data-tree
        htmlTable = "<!-- Report Generated using ZReportTool written by Adrian 'Haqa' Hungate - http://www.zope.org/Members/haqa/ZReportTool -->\n"
        htmlTable = htmlTable + '<center><table border="0" cellpadding="2" cellspacing="1" width="100%%" class="%s">\n' % self.report_class
        htmlTable = htmlTable + self._Header()
        htmlTable = htmlTable + self._DescendRows(branch=dataTree, prefix='', thisGroup='', thisValue='', groups=self.groupColumns)
        htmlTable = htmlTable + '</table></center>\n'
        return htmlTable

    def index_html(self, REQUEST=None):
        """ Call the report, and add standard_html_{header,footer} """
        htmlTable = ""
        htmlTable = htmlTable + getattr(self, "standard_html_header")(self, REQUEST)
        htmlTable = htmlTable + self(REQUEST)
        htmlTable = htmlTable + getattr(self, "standard_html_footer")(self, REQUEST)

        return htmlTable
