"""
    ZVisualQuery
    ============

A complete data query, collation, and reporting tool.

One of the main features of this product is that you do not need any
SQL programming skills - everything is visual!
"""
# $Id: ZVisualQuery.py,v 1.7 2002/01/13 21:08:58 adrian Exp $
#
# $Log: ZVisualQuery.py,v $
# Revision 1.7  2002/01/13 21:08:58  adrian
# Completed relational logic in ZDataCombiner.
# Fixed up compatibility bugs in combiner and report.
# Started online help pages.
#
# Revision 1.6  2002/01/13 09:50:24  adrian
# Fixed bug in _VisualQueryMixin that where there last character of the table
# name got lost if there was only one table in the query
#
# Revision 1.5  2002/01/13 01:22:49  adrian
# Added brain support to ODBC Socket Client
# Completed Visual Query Designer and Visual ODBC Query Designer
#
# Revision 1.4  2002/01/02 07:55:37  adrian
# Added initial support for Visual ODBC Client
#
# Revision 1.3  2001/10/29 09:29:56  adrian
# Cleaned up a number of "Ambiguous name..." warnings
#
# Revision 1.2  2001/10/21 15:03:36  adrian
# no message
#
#
__version__ = '$Revision: 1.7 $'[11:-2]

# Todo:
#  Change Output Columns page to checkbox lists
#  Change all lists to use ZMI colours

import urllib
import Globals
import string
from Products.ZSQLMethods.SQL import SQLConnectionIDs
import Shared.DC.ZRDB.DA
from Shared.DC.ZRDB.Aqueduct import parse
import ZDataCombiner
from App.Extensions import getBrain
from DocumentTemplate import DT_HTML
try: from IOBTree import Bucket
except: Bucket=lambda:{}

class _VisualQueryMixin:
    """  """
    manage_main = Globals.HTMLFile('www/queryDetails', globals())
    manage_Joins = Globals.HTMLFile('www/queryJoins', globals())
    manage_OutputColumns = Globals.HTMLFile('www/queryOutputColumns', globals())
    manage_options = (
        (
            {'label': 'Details', 'action': 'manage_main',
             'help': ('ZDataQueryKit', 'QueryDetails.stx'), },
            {'label': 'Tables', 'action': 'manage_Tables',
             'help': ('ZDataQueryKit', 'QueryTables.stx'), },
            {'label': 'Joins', 'action': 'manage_Joins',
             'help': ('ZDataQueryKit', 'QueryJoins.stx'), },
            {'label': 'Output Columns', 'action': 'manage_OutputColumns',
             'help': ('ZDataQueryKit', 'QueryColumns.stx'), },
            {'label': 'Preview SQL', 'action': 'manage_PreviewSQL',
             'help': ('ZDataQueryKit', 'QuerySQL.stx'), },
        )
    )
    relationOuterActions = (
        '=*',
        '*=',
    )
    relationActions = (
        ZDataCombiner.ZDataCombiner.relationActions +
        relationOuterActions
    )
    src = ""
    joins = []

    def __init__(self, id, title=""):
        """ Initialize this Visual Query """
        self.manage_resetQuery()
        self._clearRelLogic()

    def manage_resetQuery(self, REQUEST=None):
        """ Reset the query """
        self.tables = []
        self.columns = {}
        self.joins = []
        self.tableCache = []
        self.columnCache = []
        self._updateTemplate()
        if REQUEST:
            message = "Query reset."
            return self.manage_main(self, REQUEST, manage_tabs_message=message)

    def listTables(self):
        """ Return a list of table data in the select connection """
        raise "Override me"

    def listColumns(self, table):
        """ """
        raise "Override me"

    def manage_addTable(self, table, REQUEST=None):
        """ Add a table to the query """
        tables = self.tables
        tables.append(table)
        self.tables = tables
        clist = self.columnCache
        for column in self.listColumns(table):
            clist.append((table, column['Name']))
        self.columnCache = clist
        self._updateTemplate()
        if REQUEST:
            message = "Table added."
            return self.manage_Tables(self, REQUEST, manage_tabs_message=message)

    def manage_removeTables(self, REQUEST):
        """ Remove selected tables from the query """
        list = []
        for table in self.tables:
            if REQUEST.has_key('t_%s' % table):
                chash = self.columns
                for column in self.columns.keys():
                    if self._getTableName(column) == table:
                        del chash[column]
                self.columns = chash
                joins = []
                for item in self.joins:
                    if table not in (self._getTableName(item['leftColumn']), self._getTableName(item['rightColumn'])):
                        joins.append(item)
                self.joins = joins
                clist = []
                for column in self.columnCache:
                    if not column[0] == table:
                        clist.append(column)
                self.columnCache = clist
            else:
                list.append(table)
        self.tables = list
        self._updateTemplate()
        if REQUEST:
            message = "Table(s) removed."
            return self.manage_Tables(self, REQUEST, manage_tabs_message=message)

    def manage_recacheColumns(self, REQUEST):
        """ Reload the column lists for all the selected tables """
        clist = []
        self.columns = {}
        for table in self.tables:
            for column in self.listColumns(table):
                clist.append((table, column['Name']))
        self.columnCache = clist
        self._updateTemplate()
        if REQUEST:
            message = "Columns recached."
            return self.manage_Tables(self, REQUEST, manage_tabs_message=message)

    def manage_changeOutputColumns(self, REQUEST):
        """ """
        chash = {}
        for column in self.columnCache:
            sqlName = "%s.%s" % (column[0], column[1])
            cbFormName = "c_%s_%s" % (column[0], column[1])
            txtFormName = "a_%s_%s" % (column[0], column[1])
            if REQUEST.get(cbFormName, ''):
                chash[sqlName] = REQUEST.get(txtFormName)
        self.columns = chash
        self._updateTemplate()
        if REQUEST:
            message = "Output columns updated."
            return self.manage_OutputColumns(self, REQUEST, manage_tabs_message=message)

    def manage_addJoin(self, leftColumn, relation, rightColumn, REQUEST=None):
        """ Add a new table join """
        joins = self.joins
        joins.append({'leftColumn': str(leftColumn), 'relation': str(relation), 'rightColumn': str(rightColumn)})
        self.joins = joins
        self._updateTemplate()
        if REQUEST:
            message = "Join created."
            return self.manage_Joins(self, REQUEST, manage_tabs_message=message)

    def manage_delJoin(self, REQUEST=None):
        """ Delete one or more table joins """
        joins = self.joins
        for item in range(len(self.joins)-1, -1, -1):
            if REQUEST.get('dl_%d' % item):
                del joins[item]
        self.joins = joins
        self._updateTemplate()
        if REQUEST:
            message = "Join(s) deleted."
            return self.manage_Joins(self, REQUEST, manage_tabs_message=message)

    def _updateTemplate(self):
        """ Update the template in the underlying SQLMethod """
        self.src = str(self._generateSQL())

    def _getTableName(self, columnName):
        """ Split the table name from a FQ column name """
        if string.find(columnName, '.')==-1:
            raise "Not Fully Qualified Column Name"
        return columnName[:string.find(columnName, '.')]

    def _leftOnly(self):
        """ List the tables that appear only on the
            left hand side of a join, or appear in
            no joins at all """
        tabs = {}
        for item in self.tables:
            tabs[item] = item
        for item in self.joins:
            rtab = item['rightColumn']
            rtab = self._getTableName(rtab)
            if tabs.has_key(rtab):
                del tabs[rtab]
        if not tabs:
            raise "SQL Generation Error", "At least one table must not appear on the right"
        return tabs.keys()

    def _groupJoins(self):
        """ Group the joins for a particular 'Right Hand' Table """

    def _generateSQL(self):
        """ Generate the SQL for this Visual Query """
        columns = []
        for column in self.columns.keys():
            sc = column
            oc = self.columns[column]
            if string.find(sc, ' ')>0:
                sc = "[%s]" % sc
            if string.find(oc, ' ')>0:
                oc = "[%s]" % oc
            columns.append("%s as %s" % (sc, oc))
        cols = string.join(columns, ',\n\t')
        SQL = "SELECT"
        if hasattr(self, 'enable_max_rows_'):
            if self.enable_max_rows_ and self.max_rows_>0:
                SQL = SQL + " TOP %d" % (self.max_rows_,)
        SQL = SQL + "\n\t" + cols
# Time to find the left-only table(s) (More than one is a product)
        tabs = string.join(self._leftOnly(), ',\n\t')
        SQL = SQL + "\nFROM "
        for I in range(1, len(self.joins)):
            SQL = SQL + "("
        SQL = SQL + "\n\t" + tabs
        for join in self.joins:
            if join['relation'] not in self.relationOuterActions:
                SQL = SQL + "\nINNER JOIN %s on %s %s %s)" % (
                    self._getTableName(join['rightColumn']),
                    join['leftColumn'], join['relation'], join['rightColumn'])
        for join in self.joins:
            if join['relation'] in self.relationOuterActions:
                if join['relation'] == self.relationOuterActions[0]:
                    SQL = SQL + "\nLEFT"
                else:
                    SQL = SQL + "\nRIGHT"
                SQL = SQL + " JOIN %s on %s %s %s)" % (
                    self._getTableName(join['rightColumn']),
                    join['leftColumn'], string.replace(join['relation'], '*', ''), join['rightColumn'])
        if self.joins:
            SQL = SQL[:len(SQL)-1]
        return SQL

    def manage_PreviewSQL(self):
        """ Display the generated SQL without executing it """
        dtml = "<dtml-var manage_page_header><dtml-var manage_tabs>"
        dtml = dtml + "<h2>Your Visual Query will be executed as</h2>"
        dtml = dtml + "<blockquote><pre>%s</pre></blockquote>" % self.src
        dtml = dtml + "<dtml-var manage_page_footer>"
        obj = DT_HTML.HTML(dtml, self.REQUEST, 'ZVisualQuery - preview - %s' % self.title_and_id())
        return obj(self.aq_inner, self.REQUEST)

manage_addZVisualQueryForm = Globals.HTMLFile('www/addZVisualQuery', globals())
def manage_addZVisualQuery(self, id, title='', submit='', REQUEST=None):
    """ """
    obj = ZVisualQuery(str(id), title=str(title))
    self._setObject(obj.getId(), obj)
    if REQUEST is not None:
        try: u = self.DestinationURL()
        except: u = REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u, urllib.quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_main')
    return ''

class ZVisualQuery(
        Shared.DC.ZRDB.DA.DA,
        _VisualQueryMixin,
    ):
    """ A Visual ZSQL Query Creator/Editor """
    manage_Tables = Globals.HTMLFile('www/queryTables', globals())
    meta_type = 'Z Visual Query'
    id = 'ZVisualQuery'
    manage_options = (
        _VisualQueryMixin.manage_options +
        Shared.DC.ZRDB.DA.DA.manage_options[1:]
    )

    def __init__(self, id, title=""):
        """ """
        Shared.DC.ZRDB.DA.DA.__init__(self, id, title, '', '', '')
#        _VisualQueryMixin.__init__(self, id, title="")
# Kludge:-- Rebasing problem
        self.manage_resetQuery()
        self._clearRelLogic()

    def manage_saveChanges(self, title, connection_id, arguments, REQUEST=None):
        """ Save changes in the TOP BOX (Title, connection, arguments) """
        self.title = str(title)
        self.arguments_src = str(arguments)
        self._arg = parse(self.arguments_src)
        if self.connection_id != connection_id:
            self.connection_id = str(connection_id)
            self.manage_resetQuery()
            message = "Data Source changed. Visual Query reset"
        else:
            self._updateTemplate()
            message = "Visual Query updated."
        if REQUEST:
            return self.manage_main(self, REQUEST, manage_tabs_message=message)

    def listTables(self):
        """ Return a list of table data in the select connection """
        try:
            if not self.tableCache:
                db = getattr(self, self.connection_id)
                conn = db._v_database_connection
                self.tableCache = conn.tables(rdb=0)
        except:
            pass
        return self.tableCache

    def listColumns(self, table):
        """ Return a list of the columns in a table """
        db = getattr(self, self.connection_id)
        conn = db._v_database_connection
        list = conn.columns(table)
        return list

    def _updateTemplate(self):
        """ Update the template in the underlying SQLMethod """
        self.src=str(self._generateSQL())
        self.template=t=self.template_class(self.src)
        t.cook()
        self._v_cache={}, Bucket()

Globals.InitializeClass(ZVisualQuery)

try:
    import Products.XMLKit.XMLODBCClient
    import Products.XMLKit.XML
    import socket

    manage_addZVisualODBCClientForm = Globals.HTMLFile('www/addZVisualODBCClient', globals())
    def manage_addZVisualODBCClient(self, id, server='', port=0, dsn='', submit='', REQUEST=None):
        """ """
        obj = ZVisualODBCClient(str(id), server=str(server), port=int(port), dsn=str(dsn))
        self._setObject(obj.getId(), obj)
        if REQUEST is not None:
            try: u = self.DestinationURL()
            except: u = REQUEST['URL1']
            if submit==" Add and Edit ": u="%s/%s" % (u, urllib.quote(id))
            REQUEST.RESPONSE.redirect(u+'/manage_main')
        return ''

    class ZVisualODBCClient(
            Products.XMLKit.XMLODBCClient.XMLODBCClient,
            _VisualQueryMixin,
        ):
        """ A Visual ZSQL Query Creator/Editor """
        enable_max_rows_ = 1
        manage_Tables = Globals.HTMLFile('www/XMLqueryTables', globals())
        meta_type = 'Z Visual ODBC Client'
        title = "Visual XML ODBC Socket Server Client"
        id = 'ZVisualODBCClient'
        manage_options = (
            Products.XMLKit.XMLODBCClient.XMLODBCClient.manage_options[0:1] +
            _VisualQueryMixin.manage_options[1:] +
            Products.XMLKit.XMLODBCClient.XMLODBCClient.manage_options[1:]
        )

        def __init__(self, id, server="", port=0, dsn=''):
            """ """
            Products.XMLKit.XMLODBCClient.XMLODBCClient.__init__(self, id, server, port, dsn, '')
#            _VisualQueryMixin.__init__(self, id, title="")
# Kludge:-- Rebasing problem
            self.manage_resetQuery()
            self._clearRelLogic()
            np = []
            for item in self._properties:
                if item['id'] != 'query':
                    np.append(item)
            self._properties = tuple(np)
            return "Wibble"

        def manage_advanced(self, max_rows, cache_time,
                            class_name, class_file, direct=None,
                            REQUEST=None, zclass=''):
            """ Change Advanced Properties (Borrowed from Shared.ZRDB.DA.DA)
                The max rows, and cache controls have been removed """
            self.last_reload = long(0.0)
            # paranoid type checking
            if type(max_rows) is not type(1):
                max_rows=atoi(max_rows)
            if type(cache_time) is not type(1):
                cache_time=atoi(cache_time)            
            self.max_rows_ = max_rows
            self.reloadtimer = cache_time
            class_name=str(class_name)
            class_file=str(class_file)
            self.class_name_, self.class_file_ = class_name, class_file
            self._v_brain=getBrain(self.class_file_, self.class_name_, 1)
            self.allow_simple_one_argument_traversal=direct
            if zclass:
                for d in self.aq_acquire('_getProductRegistryData')('zclasses'):
                    if ("%s/%s" % (d.get('product'),d.get('id'))) == zclass:
                        self._zclass=d['meta_class']
                        break
            self._updateTemplate()
            if REQUEST is not None:
                m="%s advanced settings have been set" % (self.meta_type,)
                return self.manage_advancedForm(self,REQUEST,manage_tabs_message=m)

        def title(self, REQUEST=None):
            """ """
            try:
                return "Visual OSS on %s:%d %s" % (self.server, self.port, self.dsn, )
            except:
                return "Visual OSS"

        def listTables(self):
            """ Can't list tables via XML ODBC """
            return None

        def listColumns(self, table):
            """ Return a list of the columns in a table """
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.server, self.port), )
            file = sock.makefile('rb')
            Ssend = "<?xml version=\"1.0\"?>\r\n <request>\r\n<connectionstring>" 
            Ssend = Ssend + self.dsn + "</connectionstring>\r\n<sql>"
            Ssend = Ssend + "select top 1 * from %s" % table
            Ssend = Ssend + "</sql>\r\n</request>\r\n"
            sock.send(Ssend)
            retStr = file.read()
            sock.close()
            xp = Products.XMLKit.XML.XMLParser.XMLParser()
            xp.runString(retStr)
            dom = xp.DOM
            result = dom.getNextChildByName("result")
            cols = []
            if result["state"] != "success":
                error = result.getNextChildByName("error")
                raise "% - Query Failed: %s" % (self.meta_type, error.getCDATA(), )
            result = dom.getNextChildByName("result")
            cols = []
            row = result.getNextChildByName("row")
            for item in row.getChildrenByName("column"):
                cols.append({'Name': item["name"]})
            return tuple(cols)

        def _updateTemplate(self):
            """ Update the template in the underlying SQLMethod """
            self.last_reload = long(0.0)
            self.src = str(self._generateSQL())
            self.query = self.src

    Globals.InitializeClass(ZVisualODBCClient)
except:
    pass
